"""
External API of the aioclock package, that can be used to interact with the AioClock instance.
This module could be very useful if you intend to use aioclock in a web application or a CLI tool.

Other tools and extension are written from this tool.

!!! danger "Note when writing to aioclock API and changing its state."
    Right now the state of AioClock instance is on the memory level, so if you write an API and change a task's trigger time, it will not persist.
    In future we might store the state of AioClock instance in a database, so that it always remains same.
    But this is a bit tricky and implicit because then your code gets ignored and database is preferred over the database.
    For now you may consider it as a way to change something without redeploying the application, but it is not very recommended to write.
"""

import sys
from typing import Any, Awaitable, Callable, TypeVar, Union
from uuid import UUID

from fast_depends import inject
from pydantic import BaseModel

from aioclock.app import AioClock
from aioclock.exceptions import TaskIdNotFound
from aioclock.provider import get_provider
from aioclock.triggers import TriggerT

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

T = TypeVar("T")
P = ParamSpec("P")


class TaskMetadata(BaseModel):
    """Metadata of the task that is included in the AioClock instance.

    Attributes:
        id: UUID: Task ID that is unique for each task, and changes every time you run the aioclock app.
            In future we might store task ID in a database, so that it always remains same.
        trigger: Union[TriggerT, Any]: Trigger that is used to run the task, type is also any to ease implementing new triggers.
        task_name: str: Name of the task function.
    """

    id: UUID

    trigger: Union[TriggerT, Any]

    task_name: str


async def run_specific_task(task_id: UUID, app: AioClock):
    """Run a specific task immediately by its ID, from the AioClock instance.

    params:
        task_id: Task ID that is unique for each task, and changes every time you run the aioclock app.
            In future we might store task ID in a database, so that it always remains same.
        app: AioClock instance to run the task from.

    Example:
        ```python
        from aioclock import  AioClock, Once
        from aioclock.api import run_specific_task

        app = AioClock()

        @app.task(trigger=Once())
        async def main():
            print("Hello World")

        async def some_other_func():
            await run_specific_task(app._tasks[0].id, app)
        ```

    """
    task = next((task for task in app._tasks if task.id == task_id), None)
    if not task:
        raise TaskIdNotFound
    return await run_with_injected_deps(task.func)


async def run_with_injected_deps(func: Callable[P, Awaitable[T]]) -> T:
    """Runs an aioclock decorated function, with all the dependencies injected.

    Can be used to run a task function with all the dependencies injected.

    params:
        func: Function to run with all the dependencies injected. Must be decorated with `@app.task` decorator.

    Example:
        ```python
        from aioclock import Once, AioClock, Depends
        from aioclock.api import run_with_injected_deps

        app = AioClock()

        def some_dependency():
            return 1

        @app.task(trigger=Once())
        async def main(bar: int = Depends(some_dependency)):
            print("Hello World")
            return bar

        async def some_other_func():
            foo = await run_with_injected_deps(main)
            assert foo == 1
        ```

    """
    return await inject(func, dependency_overrides_provider=get_provider())()  # type: ignore


async def get_metadata_of_all_tasks(app: AioClock) -> list[TaskMetadata]:
    """Get metadata of all tasks that are included in the AioClock instance.

    This function can be used to mutate the `TaskMetadata` object, i.e to change the trigger of a task.
    But for now it is yet not recommended to do this, as you might experience some unexpected behavior.
    But in next versions, I'd like to make it more stable and reliable on mutating the data.

    params:
        app: AioClock instance to get the metadata of all tasks.

    Example:
        ```python
        from aioclock import AioClock, Once
        from aioclock.api import get_metadata_of_all_tasks

        app = AioClock()
        @app.task(trigger=Once())
        async def main(): ...

        async def some_other_func():
            metadata = await get_metadata_of_all_tasks(app)
        ```
    """
    return [
        TaskMetadata(
            id=task.id,
            trigger=task.trigger,
            task_name=task.func.__name__,
        )
        for task in app._get_tasks(exclude_type=set())
    ]



"""
To initialize the AioClock instance, you need to import the AioClock class from the aioclock module.
AioClock class represent the aioclock, and handle the tasks and groups that will be run by the aioclock.

Another way to modulize your code is to use `Group` which is kinda the same idea as router in web frameworks.
"""

import asyncio
import sys
from functools import wraps
from typing import Any, Callable, Optional, TypeVar, Union

import anyio

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

from asyncer import asyncify
from fast_depends import inject

from aioclock.custom_types import Triggers
from aioclock.group import Group, Task
from aioclock.provider import get_provider
from aioclock.triggers import BaseTrigger
from aioclock.utils import flatten_chain

T = TypeVar("T")
P = ParamSpec("P")


class AioClock:
    """
    AioClock is the main class that will be used to run the tasks.
    It will be responsible for running the tasks in the right order.

    Example:
        ```python
        from aioclock import AioClock, Once
        app = AioClock()

        @app.task(trigger=Once())
        async def main():
            print("Hello World")
        ```

    To run the aioclock final app simply do:

    Example:
        ```python
        from aioclock import AioClock, Once
        import asyncio

        app = AioClock()

        # whatever next comes here
        asyncio.run(app.serve())
        ```

    """

    def __init__(self, limiter: Optional[anyio.CapacityLimiter] = None):
        """
        Initialize AioClock instance.
        No parameters are needed.

        Attributes:
            limiter:
                Anyio CapacityLimiter. capacity limiter to use to limit the total amount of threads running
                Limiter that will be used to limit the number of tasks that are running at the same time.
                If not provided, it will fallback to the default limiter set on Application level.
                If no limiter is set on Application level, it will fallback to the default limiter set by AnyIO.

        """
        self._groups: list[Group] = []
        self._app_tasks: list[Task] = []
        self._limiter = limiter

    _groups: list[Group]
    """List of groups that will be run by AioClock."""

    _app_tasks: list[Task]
    """List of tasks that will be run by AioClock."""

    @property
    def dependencies(self):
        """Dependencies provider that will be used to inject dependencies in tasks."""
        return get_provider()

    def override_dependencies(
        self, original: Callable[..., Any], override: Callable[..., Any]
    ) -> None:
        """Override a dependency with a new one.

        params:
            original:
                Original dependency that will be overridden.
            override:
                New dependency that will override the original one.

        Example:
            ```python
            from aioclock import AioClock

            def original_dependency():
                return 1

            def new_dependency():
                return 2

            app = AioClock()
            app.override_dependencies(original=original_dependency, override=new_dependency)
            ```

        """
        self.dependencies.override(original, override)

    def include_group(self, group: Group) -> None:
        """Include a group of tasks that will be run by AioClock.

        params:
            group:
                Group of tasks that will be run together.

        Example:
            ```python
            from aioclock import AioClock, Group, Once

            app = AioClock()

            group = Group()
            @group.task(trigger=Once())
            async def main():
                print("Hello World")

            app.include_group(group)
            ```
        """
        self._groups.append(group)
        return None

    def task(self, *, trigger: BaseTrigger):
        """
        Decorator to add a task to the AioClock instance.
        If decorated function is sync, aioclock will run it in a thread pool executor, using AnyIO.
        But if you try to run the decorated function, it will run in the same thread, blocking the event loop.
        It is intended to not change all your `sync functions` to coroutine functions,
            and they can be used outside of aioclock, if needed.

        params:
            trigger: BaseTrigger
                Trigger that will trigger the task to be running.

        Example:
            ```python

            from aioclock import AioClock, Once

            app = AioClock()

            @app.task(trigger=Once())
            async def main():
                print("Hello World")
            ```
        """

        def decorator(func):
            @wraps(func)
            async def wrapped_funciton(*args, **kwargs):
                if asyncio.iscoroutinefunction(func):
                    return await func(*args, **kwargs)
                else:  # run in threadpool to make sure it's not blocking the event loop
                    return await asyncify(func, limiter=self._limiter)(*args, **kwargs)

            self._app_tasks.append(
                Task(
                    func=inject(wrapped_funciton, dependency_overrides_provider=get_provider()),
                    trigger=trigger,
                )
            )
            if asyncio.iscoroutinefunction(func):
                return wrapped_funciton
            else:

                @wraps(func)
                def wrapper(*args, **kwargs):
                    return func(*args, **kwargs)

                return wrapper

        return decorator

    @property
    def _tasks(self) -> list[Task]:
        result = flatten_chain([group._tasks for group in self._groups])
        return result

    def _get_shutdown_task(self) -> list[Task]:
        return [task for task in self._tasks if task.trigger.type_ == Triggers.ON_SHUT_DOWN]

    def _get_startup_task(self) -> list[Task]:
        return [task for task in self._tasks if task.trigger.type_ == Triggers.ON_START_UP]

    def _get_tasks(self, exclude_type: Union[set[Triggers], None] = None) -> list[Task]:
        exclude_type = (
            exclude_type
            if exclude_type is not None
            else {Triggers.ON_START_UP, Triggers.ON_SHUT_DOWN}
        )

        return [task for task in self._tasks if task.trigger.type_ not in exclude_type]

    async def serve(self) -> None:
        """
        Serves AioClock
        Run the tasks in the right order.
        First, run the startup tasks, then run the tasks, and finally run the shutdown tasks.
        """
        group = Group()
        group._tasks = self._app_tasks
        self.include_group(group)
        try:
            await asyncio.gather(
                *(task.run() for task in self._get_startup_task()), return_exceptions=False
            )

            await asyncio.gather(
                *(group.run() for group in self._get_tasks()), return_exceptions=False
            )
        finally:
            shutdown_tasks = self._get_shutdown_task()
            await asyncio.gather(*(task.run() for task in shutdown_tasks), return_exceptions=False)



import asyncio
import threading
from time import sleep
from typing import Annotated

from aioclock import AioClock, Depends, Every, Group, OnShutDown, OnStartUp

# service1.py
group = Group()


def dependency():
    return "Hello from thread: "


@group.task(trigger=Every(seconds=2))
def sync_task_1(val: str = Depends(dependency)):
    print(f"{val} `sync_task_1` {threading.current_thread().ident}")
    sleep(1)  # some blocking operation


@group.task(trigger=Every(seconds=2.01))
def sync_task_2(val: Annotated[str, Depends(dependency)]):
    print(f"{val} `sync_task_2` {threading.current_thread().ident}")
    sleep(1)  # some blocking operation
    return "3"


print(sync_task_2("Aioclock won't color your functions! "))


@group.task(trigger=Every(seconds=2))
async def async_task(val: str = Depends(dependency)):
    print(f"{val} `async_task` {threading.current_thread().ident}")


# app.py
app = AioClock()
app.include_group(group)


@app.task(trigger=OnStartUp())
def startup(val: str = Depends(dependency)):
    print("Welcome!")


@app.task(trigger=OnShutDown())
def shutdown(val: str = Depends(dependency)):
    print("Bye!")


if __name__ == "__main__":
    asyncio.run(app.serve())



import logging
import subprocess
import sys

from aioclock import __version__

logger = logging.getLogger(__name__)


def run_command(command: str) -> None:
    try:
        result = subprocess.run(command, shell=True, check=True, capture_output=True, text=True)
        print(result.stdout)
    except subprocess.CalledProcessError:
        logger.exception("Error executing command")
        sys.exit(1)


if __name__ == "__main__":
    command = "mike set-default latest"
    run_command(command)

    command = f"mike deploy --push --update-aliases {__version__} latest"
    run_command(command)



import asyncio
import sys
from functools import wraps
from typing import Optional, TypeVar

from asyncer import asyncify

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

import anyio
from fast_depends import inject

from aioclock.provider import get_provider
from aioclock.task import Task
from aioclock.triggers import BaseTrigger

T = TypeVar("T")
P = ParamSpec("P")


class Group:
    def __init__(
        self,
        *,
        limiter: Optional[anyio.CapacityLimiter] = None,
    ):
        """
        Group of tasks that will be run together.

        Best use case is to have a good modularity and separation of concerns.
        For example, you can have a group of tasks that are responsible for sending emails.
        And another group of tasks that are responsible for sending notifications.

        Params:
            limiter:
                Anyio CapacityLimiter. capacity limiter to use to limit the total amount of threads running
                Limiter that will be used to limit the number of tasks that are running at the same time.
                If not provided, it will fallback to the default limiter set on Application level.
                If no limiter is set on Application level, it will fallback to the default limiter set by AnyIO.

        Example:
            ```python

            from aioclock import Group, AioClock, Forever

            email_group = Group()

            # consider this as different file
            @email_group.task(trigger=Forever())
            async def send_email():
                ...

            # app.py
            aio_clock = AioClock()
            aio_clock.include_group(email_group)
            ```

        """
        self._tasks: list[Task] = []
        self._limiter = limiter

    def task(self, *, trigger: BaseTrigger):
        """
        Decorator to add a task to the AioClock instance.
        If decorated function is sync, aioclock will run it in a thread pool executor, using AnyIO.
        But if you try to run the decorated function, it will run in the same thread, blocking the event loop.
        It is intended to not change all your `sync functions` to coroutine functions,
            and they can be used outside of aioclock, if needed.

        params:
            trigger: BaseTrigger
                Trigger that will trigger the task to be running.

        Example:
            ```python

            from aioclock import AioClock, Once

            app = AioClock()

            @app.task(trigger=Once())
            async def main():
                print("Hello World")
            ```
        """

        def decorator(func):
            @wraps(func)
            async def wrapped_funciton(*args, **kwargs):
                if asyncio.iscoroutinefunction(func):
                    return await func(*args, **kwargs)
                else:  # run in threadpool to make sure it's not blocking the event loop
                    return await asyncify(func, limiter=self._limiter)(*args, **kwargs)

            self._tasks.append(
                Task(
                    func=inject(wrapped_funciton, dependency_overrides_provider=get_provider()),
                    trigger=trigger,
                )
            )

            if asyncio.iscoroutinefunction(func):
                return wrapped_funciton

            else:

                @wraps(func)
                def wrapper(*args, **kwargs):
                    return func(*args, **kwargs)

                return wrapper

        return decorator

    async def _run(self):
        """
        Just for purpose of being able to run all task in group
        Private method, should not be used outside of the library
        """
        await asyncio.gather(
            *(task.run() for task in self._tasks),
            return_exceptions=False,
        )



import asyncio
from contextlib import asynccontextmanager

from fastapi import FastAPI

from aioclock import AioClock
from aioclock.ext.fast import make_fastapi_router
from aioclock.triggers import Every, OnStartUp

clock_app = AioClock()


@clock_app.task(trigger=OnStartUp())
async def startup():
    print("Starting...")


@clock_app.task(trigger=Every(seconds=3600))
async def foo():
    print("Foo is processing...")


@asynccontextmanager
async def lifespan(app: FastAPI):
    task = asyncio.create_task(clock_app.serve())
    yield

    try:
        task.cancel()
        await task
    except asyncio.CancelledError:
        ...


app = FastAPI(lifespan=lifespan)
app.include_router(make_fastapi_router(clock_app))

if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app)



import asyncio

from aioclock import AioClock, At, Depends, Every, Forever, Once, OnShutDown, OnStartUp
from aioclock.group import Group

# groups.py
group = Group()


def more_useless_than_me():
    return "I'm a dependency. I'm more useless than a screen door on a submarine."


@group.task(trigger=Every(seconds=10))
async def every():
    print("Every 10 seconds, I make a quantum leap. Where will I land next?")


@group.task(trigger=At(tz="UTC", hour=0, minute=0, second=0))
async def at():
    print(
        "When the clock strikes midnight... I turn into a pumpkin. Just kidding, I run this task!"
    )


@group.task(trigger=Forever())
async def forever(val: str = Depends(more_useless_than_me)):
    await asyncio.sleep(2)
    print("Heartbeat detected. Still not a zombie. Will check again in a bit.")
    assert val == "I'm a dependency. I'm more useless than a screen door on a submarine."


@group.task(trigger=Once())
async def once():
    print("Just once, I get to say something. Here it goes... I love lamp.")


# app.py
app = AioClock()
app.include_group(group)


@app.task(trigger=OnStartUp())
async def startup():
    print(
        "Welcome to the Async Chronicles! Did you know a group of unicorns is called a blessing? Well, now you do!"
    )


@app.task(trigger=OnShutDown())
async def shutdown():
    print("Going offline. Remember, if your code is running, you better go catch it!")


# main.py
if __name__ == "__main__":
    asyncio.run(app.serve())



import asyncio

from aioclock import AioClock, Depends, Every, Group

# service1.py
group = Group()


def dependency():
    return "Hello, world!"


def overwritten_dependency():
    return "Goodbye, world!"


@group.task(trigger=Every(seconds=1))
async def my_task(val: str = Depends(dependency)):
    print(val)


# app.py
app = AioClock()
app.include_group(group)


app.override_dependencies(dependency, overwritten_dependency)

if __name__ == "__main__":
    asyncio.run(app.serve())



from enum import auto
from typing import Annotated, Literal, Union

from annotated_types import Interval

from aioclock.utils import StrEnum

EveryT = Literal[
    "every monday",
    "every tuesday",
    "every wednesday",
    "every thursday",
    "every friday",
    "every saturday",
    "every sunday",
    "every day",
]

SecondT = Annotated[int, Interval(ge=0, le=59)]
MinuteT = Annotated[int, Interval(ge=0, le=59)]
HourT = Annotated[int, Interval(ge=0, le=24)]

PositiveNumber = Annotated[Union[int, float], Interval(ge=0)]


class Triggers(StrEnum):
    # TODO: support cron job trigger
    # CRON = auto()
    # """Cron job trigger."""
    EVERY = auto()
    """Every (x) time units, it gets triggered."""
    ONCE = auto()
    """Trigger once, then stop."""
    FOREVER = auto()
    """Keep running, as long as application is running."""
    ON_START_UP = auto()
    """Trigger on application start up."""
    ON_SHUT_DOWN = auto()
    """Trigger on application shut down."""
    AT = auto()
    """Trigger at a specific time."""



"""
Aioclock wrap your functions with a task object, and append the task to the list of tasks in the AioClock instance.
After collecting all the tasks from decorated functions, aioclock serve them in order it has to be (startup, normal, shutdown).

These tasks keep running forever until the trigger's method `should_trigger` returns False.
"""

from dataclasses import dataclass, field
from datetime import datetime, timedelta, timezone
from typing import Any, Awaitable, Callable
from uuid import UUID, uuid4

from aioclock.logger import logger
from aioclock.triggers import BaseTrigger

UTC = timezone.utc


@dataclass
class Task:
    """Task that will be run by AioClock.
    Which always has a function and a trigger.
    This is internally used, when you decorate your function with `aioclock.task`.

    Attributes:
        func: Callable[..., Awaitable[Any]]: Decorated function that will be run by AioClock.
        trigger: BaseTrigger: Trigger that will be used to run the function.
        id: UUID: Task ID that is unique for each task, and changes every time you run the aioclock app.
            In future we might store task ID in a database, so that it always remains same.

    """

    func: Callable[..., Awaitable[Any]]

    trigger: BaseTrigger

    id: UUID = field(default_factory=uuid4)

    async def run(self):
        """
        Run the task, and handle the exceptions.
        If the task fails, log the error with exception, but keep running the tasks.
        """
        while self.trigger.should_trigger():
            try:
                next_trigger = await self.trigger.get_waiting_time_till_next_trigger()
                if next_trigger is not None:
                    logger.info(f"Triggering next task {self.func.__name__} in {next_trigger}")
                    self.trigger.expected_trigger_time = datetime.now(UTC) + timedelta(
                        seconds=next_trigger
                    )
                await self.trigger.trigger_next()
                logger.debug(f"Running task {self.func.__name__}")
                await self.func()
            except Exception as error:
                # Log the error, but keep running the tasks.
                # don't crash the whole application.
                logger.exception(f"Error running task {self.func.__name__}: {error}")

        self.trigger.expected_trigger_time = None



import pytest

from aioclock import AioClock, Depends, Once
from aioclock.api import run_with_injected_deps

app = AioClock()


def some_dependency():
    return 1


@app.task(trigger=Once())
async def main(bar: int = Depends(some_dependency)):
    print("Hello World")
    return bar


@pytest.mark.asyncio
async def test_run_manual():
    foo = await run_with_injected_deps(main)
    assert foo == 1



import importlib.util
import inspect
import logging
import os
import sys
import textwrap

logger = logging.getLogger(__name__)


def extract_code_blocks(docstring: str) -> list[str]:
    """Extract code blocks from a docstring."""
    code_blocks = []
    in_code_block = False
    code_block = []

    for line in docstring.split("\n"):
        if line.strip().startswith("```python"):
            in_code_block = True
            continue
        elif line.strip().startswith("```"):
            in_code_block = False
            if code_block:
                code_blocks.append("\n".join(code_block))
                code_block = []
            continue

        if in_code_block:
            code_block.append(line)

    return code_blocks


def exec_example(example: str) -> None:
    """Execute a code example."""
    example = textwrap.dedent(example)
    namespace = {"__name__": "__not_main__"}
    try:
        code = compile(example, "<string>", "exec")
        exec(code, namespace)
    except Exception:
        raise Exception(f"Failed to execute example:\n\n{example}")


def process_module(module_path: str) -> None:
    """Process a single Python module."""
    module_name = os.path.splitext(os.path.basename(module_path))[0]
    spec = importlib.util.spec_from_file_location(module_name, module_path)
    if spec is None or spec.loader is None:
        raise Exception(f"Could not load module: {module_path}")

    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

    for name, obj in inspect.getmembers(module):
        if inspect.isfunction(obj) or inspect.isclass(obj):
            docstring = inspect.getdoc(obj)
            if docstring:
                examples = extract_code_blocks(docstring)
                for example in examples:
                    exec_example(example)


def process_markdown(markdown_path: str) -> None:
    """Process a markdown file and execute code blocks."""
    with open(markdown_path, "r") as file:
        content = file.read()

    code_blocks = extract_code_blocks(content)
    for example in code_blocks:
        exec_example(example)


def traverse_library(library_path: str) -> None:
    """Traverse the library directory and process each Python module."""
    original_sys_path = sys.path.copy()
    sys.path.insert(0, os.path.abspath(library_path))  # Prioritize the library path

    for root, _, files in os.walk(library_path):
        for file in files:
            if file.endswith(".py"):
                module_path = os.path.join(root, file)
                process_module(module_path)

    sys.path = original_sys_path  # Restore the original sys.path


def traverse_docs(docs_path: str) -> None:
    """Traverse the docs directory and process each markdown file."""
    for root, _, files in os.walk(docs_path):
        for file in files:
            if file.endswith(".md"):
                markdown_path = os.path.join(root, file)
                process_markdown(markdown_path)


def test_examples():
    docs_path = "../aioclock/docs"
    traverse_docs(docs_path)

    library_path = "../aioclock/aioclock"
    traverse_library(library_path)



from functools import lru_cache

from fast_depends import Provider


@lru_cache
def get_provider():
    """Return a Provider instance, which is singleton.
    This singleton is used to inject dependencies in tasks.
    """
    return Provider()



"""
Triggers are used to determine when the event should be triggered. It can be based on time, or some other condition.
You can create custom triggers by inheriting from `BaseTrigger` class.

!!! info "Don't run CPU intensitve or thread-block IO task "
    AioClock's trigger are all running in async, only on one CPU.
    So, if you run a CPU intensive task, or a task that blocks the thread, then it will block the entire event loop.
    If you have a sync IO task, then it's recommended to use `run_in_executor` to run the task in a separate thread.
    Or use similiar libraries like `asyncer` or `trio` to run the task in a separate thread.
"""

import asyncio
from abc import ABC, abstractmethod
from copy import deepcopy
from datetime import datetime, timedelta
from typing import Annotated, Generic, Literal, TypeVar, Union

import zoneinfo
from annotated_types import Interval
from pydantic import BaseModel, Field, PositiveInt, model_validator

from aioclock.custom_types import EveryT, PositiveNumber, Triggers

TriggerTypeT = TypeVar("TriggerTypeT")


class BaseTrigger(BaseModel, ABC, Generic[TriggerTypeT]):
    """
    Base class for all triggers.
    A trigger is a way to determine when the event should be triggered. It can be based on time, or some other condition.


    The way trigger are used is as follows:
        1. An async function which is a task, is decorated with framework, and trigger is the arguement for the decorator
        2. `get_waiting_time_till_next_trigger` is called to get the time in seconds, after which the event should be triggered.
        3. If the time is not None, then it logs the time that is predicted for the event to be triggered.
        4. `trigger_next` is called immidiately after that, which triggers the event.

    You can create trigger by yourself, by inheriting from `BaseTrigger` class.

    Example:
        ```python
        from aioclock.triggers import BaseTrigger
        from typing import Literal

        class Forever(BaseTrigger[Literal["Forever"]]):
            type_: Literal["Forever"] = "Forever"

            def should_trigger(self) -> bool:
                return True

            async def trigger_next(self) -> None:
                return None

            async def get_waiting_time_till_next_trigger(self):
                if self.should_trigger():
                    return 0
                return None
        ```

    Attributes:
        type_: Type of the trigger. It is a string, which is used to identify the trigger's name.
            You can change the type by using `Generic` type when inheriting from `BaseTrigger`.
    """

    type_: TriggerTypeT
    expected_trigger_time: Union[datetime, None] = None

    @abstractmethod
    async def trigger_next(self) -> None:
        """
        `trigger_next` keep track of the event, and triggers the event.
        The function shall return when the event is triggered and should be executed.
        """

    def should_trigger(self) -> bool:
        """
        `should_trigger` checks if the event should be triggered or not.
        If not, then the event will not be triggered anymore.
        You can save the state of the trigger and task inside the instance, and then check if the event should be triggered or not.
        For instance, in `LoopCounter` trigger, it keeps track of the number of times the event has been triggered,
        and then checks if the event should be triggered or not.
        """
        return True

    @abstractmethod
    async def get_waiting_time_till_next_trigger(self) -> Union[float, None]:
        """
        Returns the time in seconds, after which the event should be triggered.
        Returns None, if the event should not trigger anymore.
        """
        ...


class Forever(BaseTrigger[Literal[Triggers.FOREVER]]):
    """A trigger that is always triggered imidiately.

    Example:
        ```python

            from aioclock import AioClock, Forever

            app = AioClock()

            # instead of this:
            async def my_task():
                while True:
                    try:
                        await asyncio.sleep(3)
                        1/0
                    except DivisionByZero:
                        pass

            # use this:
            @app.task(trigger=Forever())
            async def my_task():
                await asyncio.sleep(3)
                1/0
        ```

    Attributes:
        type_: Type of the trigger. It is a string, which is used to identify the trigger's name.
            You can change the type by using `Generic` type when inheriting from `BaseTrigger`.

    """

    type_: Literal[Triggers.FOREVER] = Triggers.FOREVER

    def should_trigger(self) -> bool:
        return True

    async def trigger_next(self) -> None:
        return None

    async def get_waiting_time_till_next_trigger(self):
        return 0


class LoopController(BaseTrigger, ABC, Generic[TriggerTypeT]):
    """
    Base class for all triggers that have loop control.

    Attributes:
        type_: Type of the trigger. It is a string, which is used to identify the trigger's name.
            You can change the type by using `Generic` type when inheriting from `LoopController`.

        max_loop_count: The maximum number of times the event should be triggered.
            If set to 3, then 4th time the event will not be triggered.
            If set to None, it will keep running forever.
            This is available for all triggers that inherit from `LoopController`.

        _current_loop_count: Current loop count, which is used to keep track of the number of times the event has been triggered.
            Private attribute, should not be accessed directly.
            This is available for all triggers that inherit from `LoopController`.
    """

    type_: TriggerTypeT
    _current_loop_count: int = 0
    max_loop_count: Union[PositiveInt, None] = None

    @model_validator(mode="after")
    def validate_loop_controll(self):
        if "_current_loop_count" in self.model_fields_set:
            raise ValueError("_current_loop_count is a private attribute, should not be provided.")
        return self

    def _increment_loop_counter(self) -> None:
        self._current_loop_count += 1

    def should_trigger(self) -> bool:
        if self.max_loop_count is None:
            return True
        if self._current_loop_count < self.max_loop_count:
            return True
        return False

    async def get_waiting_time_till_next_trigger(self):
        return 0


class Once(LoopController[Literal[Triggers.ONCE]]):
    """A trigger that is triggered only once. It is used to trigger the event only once, and then stop.

    Example:
        ```python
        from aioclock import AioClock, Once
        app = AioClock()

        app.task(trigger=Once())
        async def task():
            print("Hello World!")
        ```
    """

    type_: Literal[Triggers.ONCE] = Triggers.ONCE
    max_loop_count: Literal[1] = 1

    async def trigger_next(self) -> None:
        self._increment_loop_counter()
        return None

    async def get_waiting_time_till_next_trigger(self):
        if self._current_loop_count == 0:
            return 0
        return None


class OnStartUp(LoopController[Literal[Triggers.ON_START_UP]]):
    """Just like Once, but it triggers the event only once, when the application starts up.

    Example:
        ```python
        from aioclock import AioClock, OnStartUp
        app = AioClock()

        app.task(trigger=OnStartUp())
        async def task():
            print("Hello World!")
        ```
    """

    type_: Literal[Triggers.ON_START_UP] = Triggers.ON_START_UP
    max_loop_count: Literal[1] = 1

    async def trigger_next(self) -> None:
        self._increment_loop_counter()
        return None

    async def get_waiting_time_till_next_trigger(self):
        if self._current_loop_count == 0:
            return 0
        return None


class OnShutDown(LoopController[Literal[Triggers.ON_SHUT_DOWN]]):
    """Just like Once, but it triggers the event only once, when the application shuts down.

    Example:
        ```python
        from aioclock import AioClock, OnShutDown
        app = AioClock()

        app.task(trigger=OnShutDown())
        async def task():
            print("Hello World!")
        ```
    """

    type_: Literal[Triggers.ON_SHUT_DOWN] = Triggers.ON_SHUT_DOWN
    max_loop_count: Literal[1] = 1

    async def trigger_next(self) -> None:
        self._increment_loop_counter()
        return None

    async def get_waiting_time_till_next_trigger(self):
        if self._current_loop_count == 0:
            return 0
        return None


class Every(LoopController[Literal[Triggers.EVERY]]):
    """A trigger that is triggered every x time units.

    Example:
        ```python
        from aioclock import AioClock, Every
        app = AioClock()

        app.task(trigger=Every(seconds=3))
        async def task():
            print("Hello World!")
        ```

    Attributes:
        first_run_strategy: Strategy to use for the first run.
            If `immediate`, then the event will be triggered immediately,
                and then wait for the time to trigger the event again.
            If `wait`, then the event will wait for the time to trigger the event for the first time.

        seconds: Seconds to wait before triggering the event.
        minutes: Minutes to wait before triggering the event.
        hours: Hours to wait before triggering the event.
        days: Days to wait before triggering the event.
        weeks: Weeks to wait before triggering the event.
        max_loop_count: The maximum number of times the event should be triggered.
    """

    type_: Literal[Triggers.EVERY] = Triggers.EVERY
    first_run_strategy: Literal["immediate", "wait"] = "wait"
    seconds: Union[PositiveNumber, None] = None
    minutes: Union[PositiveNumber, None] = None
    hours: Union[PositiveNumber, None] = None
    days: Union[PositiveNumber, None] = None
    weeks: Union[PositiveNumber, None] = None
    max_loop_count: Union[PositiveInt, None] = None

    @model_validator(mode="after")
    def validate_time_units(self):
        if (
            self.seconds is None
            and self.minutes is None
            and self.hours is None
            and self.days is None
            and self.weeks is None
        ):
            raise ValueError("At least one time unit must be provided.")

        return self

    @property
    def to_seconds(self) -> float:
        result = self.seconds or 0
        if self.weeks is not None:
            result += self.weeks * 604800
        if self.days is not None:
            result += self.days * 86400
        if self.hours is not None:
            result += self.hours * 3600
        if self.minutes is not None:
            result += self.minutes * 60

        return result

    async def trigger_next(self) -> None:
        self._increment_loop_counter()
        if self._current_loop_count == 1 and self.first_run_strategy == "immediate":
            return None
        await asyncio.sleep(self.to_seconds)
        return None

    async def get_waiting_time_till_next_trigger(self):
        # not incremented yet, so the counter is 0
        if self._current_loop_count == 0 and self.first_run_strategy == "immediate":
            return 0

        if self.should_trigger():
            return self.to_seconds
        return None


WEEK_TO_SECOND = 604800


class At(LoopController[Literal[Triggers.AT]]):
    """A trigger that is triggered at a specific time.

    Example:
        ```python

        from aioclock import AioClock, At

        app = AioClock()

        @app.task(trigger=At(hour=12, minute=30, tz="Asia/Kolkata"))
        async def task():
            print("Hello World!")
        ```

    Attributes:
        second: Second to trigger the event.
        minute: Minute to trigger the event.
        hour: Hour to trigger the event.
        at: Day of week to trigger the event. You would get the in-line typing support when using the trigger.
        tz: Timezone to use for the event.
        max_loop_count: The maximum number of times the event should be triggered.

    """

    type_: Literal[Triggers.AT] = Triggers.AT
    max_loop_count: Union[PositiveInt, None] = None
    second: Annotated[int, Interval(ge=0, le=59)] = 0
    minute: Annotated[int, Interval(ge=0, le=59)] = 0
    hour: Annotated[int, Interval(ge=0, le=24)] = 0
    at: Literal[
        "every monday",
        "every tuesday",
        "every wednesday",
        "every thursday",
        "every friday",
        "every saturday",
        "every sunday",
        "every day",
    ] = "every day"
    tz: str

    @model_validator(mode="after")
    def validate_time_units(self):
        if self.second is None and self.minute is None and self.hour is None:
            raise ValueError("At least one time unit must be provided.")

        if self.tz is not None:
            try:
                zoneinfo.ZoneInfo(self.tz)
            except Exception as error:
                raise ValueError(f"Invalid timezone provided: {error}")

        return self

    def _shift_to_week(self, target_time: datetime, tz_aware_now: datetime):
        target_weekday: dict[EveryT, Union[int, None]] = {
            "every monday": 0,
            "every tuesday": 1,
            "every wednesday": 2,
            "every thursday": 3,
            "every friday": 4,
            "every saturday": 5,
            "every sunday": 6,
            "every day": None,
        }[self.at]

        if target_weekday is None:
            if target_time < tz_aware_now:
                target_time += timedelta(days=1)
            return target_time

        days_ahead = target_weekday - tz_aware_now.weekday()  # type: ignore
        if days_ahead <= 0:
            days_ahead += 7

        if self.at == "every day":
            target_time += timedelta(days=(1 if target_time < tz_aware_now else 0))
            return target_time

        # 1 second error
        error_margin = WEEK_TO_SECOND - 1
        if days_ahead == 7 and target_time.timestamp() - tz_aware_now.timestamp() < error_margin:
            # date is today, and event is about to be triggered today. so no need to shift to 7 days.
            return target_time

        return target_time + timedelta(days_ahead)

    def _get_next_ts(self, now: datetime) -> float:
        target_time = deepcopy(now).replace(
            hour=self.hour, minute=self.minute, second=self.second, microsecond=0
        )
        target_time = self._shift_to_week(target_time, now)
        return (target_time - now).total_seconds()

    def get_sleep_time(self):
        now = datetime.now(tz=zoneinfo.ZoneInfo(self.tz))
        sleep_for = self._get_next_ts(now)
        return sleep_for

    async def get_waiting_time_till_next_trigger(self):
        return self.get_sleep_time()

    async def trigger_next(self) -> None:
        self._increment_loop_counter()
        await asyncio.sleep(self.get_sleep_time())


TriggerT = Annotated[
    Union[Forever, Once, Every, At, OnStartUp, OnShutDown], Field(discriminator="type_")
]



from datetime import datetime

import pytest
import zoneinfo

from aioclock.triggers import At, Every, Forever, LoopController, Once


def test_at_trigger():
    # test this sunday
    trigger = At(at="every sunday", hour=14, minute=1, second=0, tz="Europe/Istanbul")

    val = trigger._get_next_ts(
        datetime(
            year=2024,
            month=3,
            day=31,
            hour=14,
            minute=00,
            second=0,
            tzinfo=zoneinfo.ZoneInfo("Europe/Istanbul"),
        )
    )
    assert val == 60

    # test next week
    trigger = At(at="every sunday", hour=14, second=59, tz="Europe/Istanbul")

    val = trigger._get_next_ts(
        datetime(
            year=2024,
            month=3,
            day=31,
            hour=14,
            minute=0,
            second=0,
            tzinfo=zoneinfo.ZoneInfo("Europe/Istanbul"),
        )
    )
    assert val == 59

    # test every day
    trigger = At(at="every day", hour=14, second=59, tz="Europe/Istanbul")
    val = trigger._get_next_ts(
        datetime(
            year=2024,
            month=3,
            day=31,
            hour=14,
            minute=0,
            second=0,
            tzinfo=zoneinfo.ZoneInfo("Europe/Istanbul"),
        )
    )
    assert val == 59

    # test next week
    trigger = At(at="every saturday", hour=14, second=0, tz="Europe/Istanbul")
    val = trigger._get_next_ts(
        datetime(
            year=2024,
            month=3,
            day=31,
            hour=14,
            minute=0,
            second=0,
            tzinfo=zoneinfo.ZoneInfo("Europe/Istanbul"),
        )
    )
    assert val == 518400


@pytest.mark.asyncio
async def test_loop_controller():
    # since once trigger is triggered, it should not trigger again.
    trigger = Once()
    assert trigger.should_trigger() is True
    await trigger.trigger_next()
    assert trigger.should_trigger() is False

    class IterateFiveTime(LoopController):
        type_: str = "foo"

        async def trigger_next(self) -> None:
            self._increment_loop_counter()
            return None

    trigger = IterateFiveTime(max_loop_count=5)
    for _ in range(5):
        assert trigger.should_trigger() is True
        await trigger.trigger_next()

    assert trigger.should_trigger() is False


@pytest.mark.asyncio
async def test_forever():
    trigger = Forever()
    assert trigger.should_trigger() is True
    await trigger.trigger_next()
    assert trigger.should_trigger() is True
    await trigger.trigger_next()
    assert trigger.should_trigger() is True


@pytest.mark.asyncio
async def test_every():
    # wait should always wait for the period on first run
    trigger = Every(seconds=1, first_run_strategy="wait")
    assert await trigger.get_waiting_time_till_next_trigger() == 1

    # immediate should always execute immediately, but wait for the period from second run.
    trigger = Every(seconds=1, first_run_strategy="immediate")
    assert await trigger.get_waiting_time_till_next_trigger() == 0
    trigger._increment_loop_counter()
    assert await trigger.get_waiting_time_till_next_trigger() == 1



"""FastAPI extension to manage the tasks of the AioClock instance in HTTP Layer.

Use cases:
    - Expose the tasks of the AioClock instance in an HTTP API.
    - Show to your client which task is going to be run next, and at which time.
    - Run a specific task from an HTTP API immidiately if needed.

To use FastAPI Extension, please make sure you do `pip install aioclock[fastapi]`.

"""

from typing import Union
from uuid import UUID

from aioclock.api import TaskMetadata, get_metadata_of_all_tasks, run_specific_task
from aioclock.app import AioClock
from aioclock.exceptions import TaskIdNotFound

try:
    from fastapi.exceptions import HTTPException
    from fastapi.routing import APIRouter
except ImportError:
    raise ImportError(
        "You need to install fastapi to use aioclock with FastAPI. Please run `pip install aioclock[fastapi]`"
    )


def make_fastapi_router(aioclock: AioClock, router: Union[APIRouter, None] = None):
    """Make a FastAPI router that exposes the tasks of the AioClock instance and its external python API in HTTP Layer.
    You can pass a router to this function, and have dependencies injected in the router, or any authorization logic that you want to have.


    params:
        aioclock: AioClock instance to get the tasks from.
        router: FastAPI router to add the routes to. If not provided, a new router will be created.

    Example:
        ```python
        import asyncio
        from contextlib import asynccontextmanager

        from fastapi import FastAPI

        from aioclock import AioClock
        from aioclock.ext.fast import make_fastapi_router
        from aioclock.triggers import Every, OnStartUp

        clock_app = AioClock()

        @clock_app.task(trigger=OnStartUp())
        async def startup():
            print("Starting...")

        @clock_app.task(trigger=Every(seconds=3600))
        async def foo():
            print("Foo is processing...")


        @asynccontextmanager
        async def lifespan(app: FastAPI):
            task = asyncio.create_task(clock_app.serve())
            yield

            try:
                task.cancel()
                await task
            except asyncio.CancelledError:
                ...


        app = FastAPI(lifespan=lifespan)
        app.include_router(make_fastapi_router(clock_app))

        if __name__ == "__main__":
            import uvicorn
            # uvicorn.run(app)
        ```
    """
    router = router or APIRouter()

    @router.get("/tasks")
    async def get_tasks() -> list[TaskMetadata]:
        return await get_metadata_of_all_tasks(aioclock)

    @router.post("/task/{task_id}")
    async def run_task(task_id: UUID):
        try:
            await run_specific_task(task_id, aioclock)
        except TaskIdNotFound:
            raise HTTPException(status_code=404, detail="Task not found")

    return router



"""
Extensions for aioclock.

AioClock is very extensible, and you can add your own extensions to it.
The extension would allow you to interact with your AioClock instance, from different layers of your application.
For instance, the FastAPI plugin allows you to run a specific task immediately from an HTTP API, or see your tasks in an HTTP API, and when they are going to run next.
"""
